import os
import time
#import numpy as np

import logging

LOG_FILENAME = '__logfile.txt'

try:
    os.remove(LOG_FILENAME)
except:
    pass
    
logging.basicConfig(filename=LOG_FILENAME, filemode='w',level=logging.DEBUG)

# www.pythonware.com/library/pil/handbook
from PIL import Image,ImageFont, ImageDraw
from PIL import ImageEnhance, ImageFilter

import Animate.Scripts
import Animate.Constants

#Todo: make this platform and -system independent
DEFAULTFONT = "C:\\WINDOWS\\Fonts\\verdana.TTF"

def Model(FileName, SheetName):

    print('### PARSING SCRIPTS  ###')
    logging.debug('### PARSING SCRIPTS  ###')

    # Create general list of scripts
    ScriptList = {}

    # Read main script
    Main = Animate.Scripts.Script(FileName, SheetName, ScriptList)

    # Parse the main script
    Main.ParseScript(FileName, SheetName)

    # Add the main script to the global list of scripts
    ScriptList[SheetName]=Main

    # Do some standard checks after parsing
    print('')
    print('### CHECKING SCRIPTS  ###')
    logging.debug('### CHECKING SCRIPTS  ###')
    OK = True
    for Script in ScriptList.values():

        if not Script.StandardChecks():
            OK=False

    assert OK

    # Get the global variables from the main script
    FramesPerSecond = Main.FramesPerSecond

    FirstFrame = int(Main.FirstFrame)

    if Main.LastFrame==-1:
        LastFrame = int(Main.MaxTime*FramesPerSecond)
    else:
        LastFrame = int(Main.LastFrame)

    MaxTime = 1.0*LastFrame/FramesPerSecond

    # Deploy all scripts, items and properties
    for Script in ScriptList.values():
        Script.Deploy(MaxTime)

    # Try to store the frames in a directory called 'Frames'
    Path = os.getcwd()
    p = os.path.join(Path,'Frames')
    if os.path.exists(p):
        if os.path.isdir(p):
            Path=p
    else:
        os.mkdir(p)
        Path=p

    # Start creating frames
    print('')
    print('### GENERATING IMAGES ###')
    logging.debug('### GENERATING IMAGES  ###')
    StartTime = time.time()

    ## ToDo: Creating GIFs is needs to be made configurable
    animatedImages = []
    #animatedSize = ( 800, int(800*Main.Height/Main.Width) )
    animatedSize = ( Main.Width, Main.Height )
    

    for Frame in range(FirstFrame, LastFrame):

        FileName='Frame%05d.png' % Frame
        
        if LastFrame != FirstFrame:
            PercentageReady = 100.0 * (Frame - FirstFrame + 1) / (LastFrame - FirstFrame)
        else:
            PercentageReady = 100.0

        # Check out the right time corresponding to this frame
        Time = 1.0*Frame/FramesPerSecond
        
        SecondsPassed = time.time() - StartTime
        SecondsToGo = SecondsPassed*(100-PercentageReady)/(PercentageReady)

        print('Generating frame %05d  t=%.2f s  %.1f%% ready. %.1f seconds to go' % (Frame, Time, PercentageReady, SecondsToGo))

        # Start with a white image
        Picture = Image.new("RGBA", (Main.Width,Main.Height), color = '#FFFFFF' )

        # Get the picture from the Main script
        ItemPicture = Main.GetPicture(Time, Frame)

        # Draw the picture over the background
        Picture.paste( ItemPicture, (0,0) )

        # Show the time if desired
        if Main.ShowTime:
            Verdana = ImageFont.truetype(DEFAULTFONT, 20)
            Draw = ImageDraw.Draw(Picture)
            Draw.text( (2,Main.Height-24), '%.3f s' % Time, fill="Black", font=Verdana)

        # enhance the quality of the image
        Picture = Picture.filter(ImageFilter.SMOOTH_MORE)
        Picture = Picture.filter(ImageFilter.SHARPEN)
        logging.debug('  Frame %s saved' % FileName)
        Picture.save( os.path.join(Path, FileName), "PNG")

        animatedImages.append( Picture.resize(animatedSize, Image.ANTIALIAS) )

    if Main.AnimatedGIF:
        print("Creating animated GIF {gif}".format(gif=Main.AnimatedGIF))
        animatedImages[0].save(Main.AnimatedGIF,
                  save_all=True, append_images=animatedImages[1:], optimize=True, duration=40, loop=0)

    if Main.Movie:
        try:
            print("Creating movie {movie}".format(movie=Main.Movie))
            import ffmpeg
            stream = ffmpeg.input('Frames\Frame%05d.png')
            stream = ffmpeg.output(stream, Main.Movie)
            stream.overwrite_output()
            ffmpeg.run(stream)
        except Exception as e:
            print('Problem when creating the movie, perhaps ffmpeg is not installed')
            print('see https://ffmpeg.org/download.html')
            print(e)




import math

# Item types
IT_BASE     = "BAS"
IT_IMAGE    = "IMG"
IT_CANVAS   = "CNV"
IT_MASK     = "MSK"
IT_TEXT     = "TXT"
IT_ASSY     = "ASS"

# Commands
ScriptCommands         = ['WIDTH', 'HEIGHT', 'FRAMESPERSECOND', 'FIRSTFRAME', 'LASTFRAME', 'SHOWTIME', 'HIDETIME', 
                          'TIMEOFFSET', 'TABLE', 'SCRIPT']
AnalogProperties       = ['XPOS', 'YPOS', 'XPOLE', 'YPOLE', 'XSCALE', 'YSCALE', 'ROTATION', 'OPACITY']
MoveCommands           = ['XMOVE', 'YMOVE', 'SXMOVE', 'SYMOVE', 'RMOVE']
ZorderCommands         = ['BRINGTOFRONT', 'SENDTOBACK']
AnalogTextProperties   = ['TEXTSIZE']
DiscreteTextProperties = ['TEXTCOLOR', 'FONT']
DiscreteProperties     = ['ASSEMBLY', 'IMAGE', 'CANVAS', 'MASK', 'TEXT']

# Default font
DEFAULTFONT = "C:\\WINDOWS\\Fonts\\verdana.TTF"
#DEFAULTFONT = "Verdana.ttf"

    
# BringToFront or SendToBack
FRONT = 0
BACK  = 1

# Interpolation of position and rotation
LINEAR  = 0
CYCLOID = 1
SPRING  = 2
CLICK   = 3
ACCEL   = 4
DAMPED  = 5

CheckMove = {   'LINEAR' :    LINEAR,
                'CYCLOID':    CYCLOID,
                'SPRING' :    SPRING,
                'CLICK'  :    CLICK,
                'ACCEL'  :    ACCEL,
                'DAMPED' :    DAMPED  }

# Types and default values for all properties
DISCRETE = 0
ANALOG   = 1

DEFAULTVALUES = [   ('XPOS'      , ANALOG  , 0             ),
                    ('XMOVE'     , DISCRETE, LINEAR        ),
                    ('XSCALE'    , ANALOG  , 1             ),
                    ('SXMOVE'    , DISCRETE, LINEAR        ),
                    ('YPOS'      , ANALOG  , 0             ),
                    ('YMOVE'     , DISCRETE, LINEAR        ),
                    ('YSCALE'    , ANALOG  , 1             ),
                    ('SYMOVE'    , DISCRETE, LINEAR        ),
                    ('ROTATION'  , ANALOG  , 0             ),
                    ('RMOVE'     , DISCRETE, LINEAR        ),
                    ('XPOLE'     , ANALOG  , 0             ),
                    ('YPOLE'     , ANALOG  , 0             ),
                    ('OPACITY'   , ANALOG  , 1             ),
                    ('TIMEOFFSET', ANALOG  , 0             ),
                    ('IMAGE'     , DISCRETE, ''            ),
                    ('CANVAS'    , DISCRETE, ''            ),
                    ('MASK'      , DISCRETE, ''            ),
                    ('TEXT'      , DISCRETE, ''            ),
                    ('TEXTCOLOR' , DISCRETE, 'BLACK'       ),
                    ('TEXTSIZE'  , ANALOG  , 12            ),
                    ('FONT'      , DISCRETE, 'verdana.ttf' ),
                    ('SCRIPT'    , DISCRETE, ''            ) ]

def Displacement(interpolation, x):

    if interpolation == LINEAR:
        return x

    elif interpolation == CYCLOID:
        return x - 1/(2*math.pi)*math.sin(2*math.pi*x)
        
    elif interpolation == SPRING:
        return (441.8214882*x**8 - 1703.935071*x**7 + 2530.182847*x**6 - 
            1771.030096*x**5 + 561.4457127*x**4 - 62.98225215*x**3 + 
            5.597813989*x**2 - 0.09284320409*x - 0.000440360649)
            
    elif interpolation == CLICK:
        return x-0.4052443*math.sin(2*math.pi*x)+0.1453085*math.sin(4*math.pi*x)

    elif interpolation == ACCEL:
        return x**2

    elif interpolation == DAMPED:
        return 1-(1-x)**2






from Animate.Constants import *

class DiscreteProperty(object):

    def __init__(self):
        self.Sequence=[]

    def Append(self, time, value):
        self.Sequence.append( (time, value) )

    def Deploy(self, MaxTime, DefaultValue):

        # Sort the list
        self.Sequence.sort()

        # Insert a point if no points are present
        if len(self.Sequence)==0:
            self.Sequence.append( (0, DefaultValue) )

        # Insert a point at t=0 if needed
        if self.Sequence[0][0]>0:
            FirstValue = self.Sequence[0][1]
            self.Sequence.insert( 0, (0, FirstValue) )

        # Determine the last value in the list (for FRAMESPERSECOND etc.)
        self.LastValue = self.Sequence[-1][1]

        # Append a point at t=MaxTime if needed
        if self.Sequence[-1][0]<MaxTime:
            self.Sequence.append( (MaxTime, self.LastValue) )
            

    def Value(self, time):

        # requested time is before the start of the sequence
        t0, y0 = self.Sequence[0]
        if time<t0:
            return y0
        
        # requested time is after the end of the sequence
        t1, y1 = self.Sequence[-1]
        if t1<time:
            return y1
        
        # requested time is somewhere in between
        for t1,y1 in self.Sequence[1:]:
            
            if t0<=time<=t1:
                return y0
            
            t0, y0 = t1, y1
            
        return 0


class AnalogProperty(DiscreteProperty):

    def Value(self, time, interpolation):

        # requested time is before the start of the sequence
        t0, y0 = self.Sequence[0]
        if time<=t0:
            return y0
        
        # requested time is after the end of the sequence
        t1, y1 = self.Sequence[-1]
        if time>=t1:
            return y1
        
        # requested time is somewhere in between
        for t1,y1 in self.Sequence[1:]:

            if t0<=time<=t1:
                try:
                    relativevalue = Displacement(interpolation, (time-t0) / (t1-t0) )
                    result=y0+relativevalue*(y1-y0)
                except:
                    print("Error in properties line 76")
                    print("interpolation={inter} time={tim} t0={t0} t1={t1}".format(inter=interpolation, tim=time, t0=t0, t1=t1))
                    print("y1={y1} relative={rel} y0={y0}".format(y1=y1, rel=relativevalue, y0=y0))
                    exit(0)
                return result
            
            t0=t1
            y0=y1
        
        # No values in sequence
        return 0


# Item types
IT_ASSEMBLY = "ASS"
IT_IMAGE    = "IMG"
IT_MASK     = "MSK"
IT_TEXT     = "TXT"

class Frame():
    def __init__(self):
        self.ItemName = ''
        self.Time = 0.000
        self.ItemType = IT_IMAGE
        self.Image = '' # Used as filename or as text to be displayed
        self.Font = 'Verdana.ttf'
        self.Xpos = 0
        self.Ypos = 0
        self.Xscale = 1.0
        self.Yscale = 1.0
        self.Rotation = 0
        self.Xpole = 0
        self.Ypole = 0
        self.TextSize = 0
        self.TextColor = 'BLACK'        
        self.Opacity = 1
        self.ZOrder = 0

def Zsorter(Frame1, Frame2):
    # Sort frames, based on Z order
    if Frame1.ZOrder<Frame2.ZOrder:
        return 1
    else:
        if Frame1.ZOrder>Frame2.ZOrder:
            return -1
        else:
            return 0




from PIL import Image                         # www.pythonware.com/library/pil/handbook
import math

import logging

from Animate.Properties import *
from Animate.Constants import *

'''
The different types of item are deliberately no subclasses of Item,
because at the time of creation, the type of item may not be known.
For instance, if first XPOS of an item is given, and then IMAGE.

Therefore, all items are the same and ItemType determines the type of
item. ItemType can be changed after creation without ruining the
properties already parsed.
'''



class Item(object):
    ''' Each distinct object of the animation is stored as an item.

        The item class defines all changes on every property for one item.
        Every property has its own timeline, independent of the timelines
        of the other properties.
        The timeline is a list of (time, property) tuples, which are lateron
        transferred to discrete, equidistant frames '''

    def __init__(self):
        self.ItemName = ''
        self.ItemType = IT_BASE
        self.SheetName = ''
        
        self.TimeOffsetUsed = False

        self.Properties = {}

        # Create all properties
        for name, kind, default in DEFAULTVALUES:

            if kind == DISCRETE:
                self.Properties[name] = DiscreteProperty()
            else:
                self.Properties[name] = AnalogProperty()

        self.PrevImageName = ''
        self.LoadedImage = None


    def StandardChecks(self):
        self.TimeOffsetUsed = (len(self.Properties['TIMEOFFSET'].Sequence)>0)
        
        # To do: more checks can be done, such as checking of images exist
        
    def AddImage(self, time, filename):

        self.ItemType = IT_IMAGE
        self.Properties['IMAGE'].Append(time, filename)

    def AddMask(self, time, filename):

        self.ItemType = IT_MASK
        self.Properties['MASK'].Append(time, filename)

    def AddText(self, time, text):

        self.ItemType = IT_TEXT
        self.Properties['TEXT'].Append(time, text)

    def AddScript(self, time, script):

        self.ItemType = IT_ASSY
        self.Properties['SCRIPT'].Append(time, script)

    def AddCanvas(self, time, script):

        self.ItemType = IT_CANVAS
        self.Properties['SCRIPT'].Append(time, script)

    def Deploy(self, MaxTime):

        # Prepare each property for first use
        for name, kind, default in DEFAULTVALUES:
            self.Properties[name].Deploy(MaxTime, default)


class ItemDict(dict):
    def __init__(self, *args):
        dict.__init__(self, args)

    def __getitem__(self, key):

        if key not in self.keys():
            newitem = Item()
            newitem.ItemName = key

            dict.__setitem__(self, key, newitem)
            
        val = dict.__getitem__(self, key)
        return val

    def __setitem__(self, key, val):

        dict.__setitem__(self, key, val)



from Tkinter import Tk
import os
import sys
from os.path import join, isdir

''' 
This file retrieves all file and directory names from the
current directory and copies the result to the clipboard
of the operating system.
'''

print sys.argv[0]

# List for temporarily storing the results
filelist = []

# Sort order:
# 1 = directories
# 2 = png files
# 3 = other files

# Store the path we are looking in
pth = os.path.dirname(sys.argv[0])

for fil in os.listdir(pth):

    fullname = os.path.join(pth,fil)
    
    if os.path.isdir(fullname):
        filelist.append( (1, fil) )
    else:
        if fil[-3:]=='png':
            filelist.append( (2, fil) )
        else:
            filelist.append( (3, fil) )

# Sort the files and directories
filelist.sort()

# Assemble the string with sorted files and directories
s=''
for isd, fil in filelist:
    print fil
    s+='%s\n' % fil

# Copy the result to the cliboard
r = Tk()
r.withdraw()
r.clipboard_clear()
r.clipboard_append(s)
r.destroy()

# Report the result
print 'The following lines are copied to the clipboard:'
print s



import os
import time

# www.lexicon.net/sjmachin/xlrd.html
import xlrd

import logging

from PIL import Image # www.pythonware.com/library/pil/handbook
from PIL import ImageFont, ImageDraw, ImageEnhance
from PIL import ImageFilter

from Animate.Items import *
from Animate.Properties import *
from Animate.Constants import *

LOG_FILENAME = '__logfile.txt'
logging.basicConfig(filename=LOG_FILENAME,level=logging.DEBUG)

def SelectFont(Directories, Fonts):
    
    for Font in Fonts:
        for Path in Directories:
            try:
                FontName=os.path.join(Path,Font)
                SelectedFont = ImageFont.truetype(FontName, 20)
                return FontName
            except:
                logging.debug('%s not successful' % FontName )
            
    print('All attempts to load fonts failed')



class Script():

    def __init__(self, FileName, SheetName, ScriptList):

        logging.debug('  Script.__init__(%s, %s)' % (FileName, SheetName) )

        self.FileName        = FileName
        self.SheetName       = SheetName
        self.ScriptList      = ScriptList
        self.IsCanvas        = False
        self.FirstImage      = True
        self.ImageDir        = 'Pictures'
        
        self.FirstFrame      = 0            # Allows the processing of a subset of frames
        self.LastFrame       = -1
        self.FramesPerSecond = 10

        self.ShowTime        = False        # Display the time in each frame

        self.Movie           = False        # Can be overridden by filename of movie
        self.AnimatedGIF     = False        # Can be overridden by filename of animated gif

        self.MaxTime         = 0            # Largest time, retrieved from the parser
        
        self.TimeOffset      = 0.0          # Script, assembly or canvas can be run with an offset to the global time
        
        self.Width           = 800          # Width of the output image
        self.Height          = 600          # Height of the output image

        self.Items           = ItemDict()   # Dictionary of items

        # List of (time, item, back/front) tuples
        self.Zbuffer         = []
        self.ZbufferIndex    = 0

        # List of Items
        self.Zorder          = []

        # Picture that was processed last
        self.Picture         = False
        self.PictureFrame    = -1


    def ParseScript(self, FileName, SheetName):

        logging.debug('  Script.ParseScript(%s, %s)' % (FileName, SheetName))

        # Open excel file with frame data
        wb = xlrd.open_workbook(FileName)
        sh = wb.sheet_by_name(SheetName)

        print(' - parsing script %s' % SheetName)
        for Row in range(sh.nrows):

            # A row contains frame data if the first cell contains a float
            if sh.cell(rowx=Row,colx=0).ctype==xlrd.XL_CELL_NUMBER:

                time    = sh.cell(rowx=Row,colx=0).value
                command = sh.cell(rowx=Row,colx=1).value.upper().strip()

                if self.MaxTime<time: self.MaxTime=time

                if command == 'WIDTH':
                    # Determine the width of the output frames
                    assert sh.cell(rowx=Row,colx=2).ctype==xlrd.XL_CELL_NUMBER, \
                        "%s at row %d of sheet %s expects a number" % (command, Row+1, SheetName)

                    self.Width = int(sh.cell(rowx=Row,colx=2).value)

                elif command == 'HEIGHT':
                    # Determine the height of the output frames
                    assert sh.cell(rowx=Row,colx=2).ctype==xlrd.XL_CELL_NUMBER, \
                        "%s at row %d of sheet %s expects a number" % (command, Row+1, SheetName)

                    self.Height = int(sh.cell(rowx=Row,colx=2).value)

                elif command == 'FRAMESPERSECOND':
                    # Sets the number of frames per second for the whole movie
                    assert sh.cell(rowx=Row,colx=2).ctype==xlrd.XL_CELL_NUMBER, \
                        "%s at row %d of sheet %s expects a number in column C" % (command, Row+1, SheetName)

                    self.FramesPerSecond = sh.cell(rowx=Row,colx=2).value

                elif command == 'FIRSTFRAME':
                    # Determine the first frame to be processed,
                    # if not all frames must be processed. For debugging
                    assert sh.cell(rowx=Row,colx=2).ctype==xlrd.XL_CELL_NUMBER, \
                        "%s at row %d of sheet %s expects a number in column C" % (command, Row+1, SheetName)

                    self.FirstFrame = sh.cell(rowx=Row,colx=2).value

                elif command == 'LASTFRAME':
                    # Determine the last frame to be processed,
                    # if not all frames must be processed. For debugging
                    assert sh.cell(rowx=Row,colx=2).ctype==xlrd.XL_CELL_NUMBER, \
                        "%s at row %d of sheet %s expects a number in column C" % (command, Row+1, SheetName)

                    self.LastFrame = sh.cell(rowx=Row,colx=2).value

                elif command == 'SHOWTIME':
                    # Write the time in the lower left corner of the frames, for debug purposes
                    self.ShowTime = True

                elif command == 'HIDETIME':
                    # Do not write the time
                    self.ShowTime = False

                elif command == 'MOVIE':
                    # Sets the number of frames per second for the whole movie
                    assert sh.cell(rowx=Row,colx=2).ctype==xlrd.XL_CELL_TEXT, \
                        "%s at row %d of sheet %s expects a filename for the movie" % (command, Row+1, SheetName)

                    self.Movie= sh.cell(rowx=Row,colx=2).value
                    print("Movie {movie} will be created after generating the frames".format(movie=self.Movie))

                elif command == 'ANIMATEDGIF':
                    # Sets the number of frames per second for the whole movie
                    assert sh.cell(rowx=Row,colx=2).ctype==xlrd.XL_CELL_TEXT, \
                        "%s at row %d of sheet %s expects a filename for the animated gif" % (command, Row+1, SheetName)

                    self.AnimatedGIF= sh.cell(rowx=Row,colx=2).value
                    print("Animated GIF {gif} will be created after generating the frames".format(gif=self.AnimatedGIF))

                elif command == 'TABLE':
                    # Do not create a new script object, but import the commands in the current script
                    assert sh.cell(rowx=Row,colx=2).ctype==xlrd.XL_CELL_TEXT, \
                        "%s at row %d of sheet %s expects a string for the table name" % (command, Row+1, SheetName)

                    sheetname = sh.cell(rowx=Row,colx=2).value.strip()
                    self.ParseTable(self.FileName, sheetname)

                elif command == 'SCRIPT':
                    # Do not create a new script object, but import the commands in the current script

                    assert sh.cell(rowx=Row,colx=2).ctype==xlrd.XL_CELL_TEXT, \
                        "%s at row %d of sheet %s expects a string for the script name" % (command, Row+1, SheetName)

                    sheetname = sh.cell(rowx=Row,colx=2).value.strip()

                    self.ParseScript(self.FileName, sheetname)

                elif command == 'ASSEMBLY':
                    # Create a new script object and use the image created by this
                    # script as feed for this item
                    assert sh.cell(rowx=Row,colx=2).ctype==xlrd.XL_CELL_TEXT, \
                        "%s at row %d of sheet %s expects a string for the assembly name" % (command, Row+1, SheetName)

                    assert sh.cell(rowx=Row,colx=3).ctype==xlrd.XL_CELL_TEXT, \
                        "%s at row %d of sheet %s expects a string for the sheet name" % (command, Row+1, SheetName)

                    itemname = sh.cell(rowx=Row,colx=2).value.upper().strip()
                    sheetname = sh.cell(rowx=Row,colx=3).value

                    # If the script is not yet in the list, create it
                    if not sheetname in self.ScriptList:
                        NewScript = Script(FileName, sheetname, self.ScriptList)
                        self.ScriptList[sheetname] = NewScript
                        NewScript.ParseScript(FileName, sheetname)

                    # Assign the script to the
                    # ToDo: Implement item type directly
                    # ToDo: Implement change of script as function of time
                    self.Items[itemname].AddScript( time, sheetname )


                elif command == 'CANVAS':
                    # A canvas is an assembly of which the background is not reset for a new frame

                    assert sh.cell(rowx=Row,colx=2).ctype==xlrd.XL_CELL_TEXT, \
                        "%s at row %d of sheet %s expects a string for the item tag" % (command, Row+1, SheetName)

                    assert sh.cell(rowx=Row,colx=3).ctype==xlrd.XL_CELL_TEXT, \
                        "%s at row %d of sheet %s expects a string for the sheet name" % (command, Row+1, SheetName)

                    itemname = sh.cell(rowx=Row,colx=2).value.upper().strip()
                    sheetname = sh.cell(rowx=Row,colx=3).value

                    # If the script is not yet in the list, create it
                    if not sheetname in self.ScriptList:
                        NewScript = Script(FileName, sheetname, self.ScriptList)
                        NewScript.IsCanvas = True
                        self.ScriptList[sheetname] = NewScript
                        NewScript.ParseScript(FileName, sheetname)

                    # Assign the script to the
                    # ToDo: Implement item type directly
                    # ToDo: Implement change of script as function of time
                    self.Items[itemname].AddCanvas( time, sheetname )

                elif command == 'IMAGE':
                    # Assign a new filename for an image item
                    assert sh.cell(rowx=Row,colx=2).ctype==xlrd.XL_CELL_TEXT, \
                        "%s at row %d of sheet %s expects a string for the item tag" % (command, Row+1, SheetName)

#                    assert sh.cell(rowx=Row,colx=3).ctype==xlrd.XL_CELL_TEXT, \
#                        "%s at row %d expects a string for the filename" % (command, Row+1)

                    itemname = sh.cell(rowx=Row,colx=2).value.upper().strip()
                    filename = os.path.join(self.ImageDir, sh.cell(rowx=Row,colx=3).value)

                    assert os.path.isfile(filename), \
                        "%s at row %d could not find file %s" % (command, Row+1, filename)

                    self.Items[itemname].AddImage( time, filename )


                elif command == 'MASK':

                    # Assign a new filename for a mask item
                    assert sh.cell(rowx=Row,colx=2).ctype==xlrd.XL_CELL_TEXT, \
                        "%s at row %d of sheet %s expects a string for the item tag" % (command, Row+1, SheetName)

                    assert sh.cell(rowx=Row,colx=3).ctype==xlrd.XL_CELL_TEXT, \
                        "%s at row %d of sheet %s expects a string for the filename" % (command, Row+1, SheetName)

                    itemname = sh.cell(rowx=Row,colx=2).value.upper().strip()
                    filename = os.path.join(self.ImageDir, sh.cell(rowx=Row,colx=3).value)

                    assert os.path.isfile(filename), \
                        "%s at row %d could not find file %s" % (command, Row+1, filename)

                    self.Items[itemname].AddMask( time, filename )


                elif command == 'TEXT':
                    # Assign a new title for a text item
                    assert sh.cell(rowx=Row,colx=2).ctype==xlrd.XL_CELL_TEXT, \
                        "%s at row %d of sheet %s expects a string in column C" % (command, Row+1, SheetName)

                    assert sh.cell(rowx=Row,colx=3).ctype==xlrd.XL_CELL_TEXT, \
                        "%s at row %d of sheet %s expects a string in column D" % (command, Row+1, SheetName)

                    itemname = sh.cell(rowx=Row,colx=2).value.upper().strip()
                    title = sh.cell(rowx=Row,colx=3).value

                    self.Items[itemname].AddText( time, title )


                elif command in ['XPOS', 'YPOS', 'XPOLE', 'YPOLE', 'XSCALE', 'YSCALE', 'ROTATION', 
                                 'TIMEOFFSET', 'TEXTSIZE', 'OPACITY']:

                    # Set a new x position
                    assert sh.cell(rowx=Row,colx=2).ctype==xlrd.XL_CELL_TEXT, \
                        "%s at row %d of sheet %s expects an item name in column C" % (command, Row+1, SheetName)

                    assert sh.cell(rowx=Row,colx=3).ctype==xlrd.XL_CELL_NUMBER, \
                        "%s at row %d of sheet %s expects a number in column D" % (command, Row+1, SheetName)

                    itemname = sh.cell(rowx=Row,colx=2).value.upper().strip()
                    value = sh.cell(rowx=Row,colx=3).value

                    self.Items[itemname].Properties[command].Append(time, value)


                elif command in ['XMOVE', 'YMOVE', 'SXMOVE', 'SYMOVE', 'RMOVE']:

                    # Determine linear or cycloid movement
                    assert sh.cell(rowx=Row,colx=2).ctype==xlrd.XL_CELL_TEXT, \
                        "%s at row %d of sheet %s expects a string in column C" % (command, Row+1, SheetName)

                    assert sh.cell(rowx=Row,colx=3).ctype==xlrd.XL_CELL_TEXT, \
                        "%s at row %d of sheet %s expects a string in column D" % (command, Row+1, SheetName)

                    itemname = sh.cell(rowx=Row,colx=2).value.upper().strip()
                    move = sh.cell(rowx=Row,colx=3).value.strip().upper()

                    if move in CheckMove:
                        self.Items[itemname].Properties[command].Append(time, CheckMove[move])
                    else:
                        print("Did not recognize type of movement on row %d." % (Row+1))

                elif command in ['TEXTCOLOR', 'FONT']:
                    # Set a new text color
                    assert sh.cell(rowx=Row,colx=2).ctype==xlrd.XL_CELL_TEXT, \
                        "%s at row %d of sheet %s expects a string in column C" % (command, Row+1, SheetName)

                    assert sh.cell(rowx=Row,colx=3).ctype==xlrd.XL_CELL_TEXT, \
                        "%s at row %d of sheet %s expects a string in column D" % (command, Row+1, SheetName)

                    itemname = sh.cell(rowx=Row,colx=2).value.upper().strip()
                    textcolor = sh.cell(rowx=Row,colx=3).value.strip()
                    self.Items[itemname].Properties[command].Append(time, textcolor)

                elif command == 'BRINGTOFRONT':
                    # Bring the item to front at this time position
                    assert sh.cell(rowx=Row,colx=2).ctype==xlrd.XL_CELL_TEXT, \
                        "%s at row %d of sheet %s expects a string in column C" % (command, Row+1, SheetName)

                    itemname = sh.cell(rowx=Row,colx=2).value.upper().strip()

                    self.Zbuffer.append( ( time, itemname, FRONT) )

                elif command == 'SENDTOBACK':
                    # Send the item to back at this time position
                    assert sh.cell(rowx=Row,colx=2).ctype==xlrd.XL_CELL_TEXT, \
                        "%s at row %d of sheet %s expects a string in column C" % (command, Row+1, SheetName)

                    itemname = sh.cell(rowx=Row,colx=2).value.upper().strip()
                    self.Zbuffer.append( ( time, itemname, BACK) )

                else:
                    print("Command %s not recognized on row %d." % (command, Row+1))

    def ParseTable(self, FileName, SheetName):

        logging.debug('  Script.ParseTable(%s, %s)' % (FileName, SheetName))

        # Open excel file with frame data
        wb = xlrd.open_workbook(FileName)
        sh = wb.sheet_by_name(SheetName)

        # Investigate which data each column contains
        print(' - parsing table %s' % SheetName)

        for Row in range(2, sh.nrows):

            # Only process rows with a time in the first column
            if sh.cell(rowx=Row,colx=0).ctype==xlrd.XL_CELL_NUMBER:

                time = sh.cell(rowx=Row,colx=0).value

                if self.MaxTime<time: self.MaxTime=time

                for Col in range(1, sh.ncols):

                    # Only process columns with an existing object in the first row and a command in the second row
                    if     sh.cell(rowx=0,colx=Col).ctype==xlrd.XL_CELL_TEXT and \
                        sh.cell(rowx=1,colx=Col).ctype==xlrd.XL_CELL_TEXT and \
                        len(sh.cell(rowx=0,colx=Col).value)>0 and \
                        len(sh.cell(rowx=1,colx=Col).value)>0:

                        itemname = sh.cell(rowx=0,colx=Col).value.upper().strip()
                        command = sh.cell(rowx=1,colx=Col).value.upper().strip()

                        # Only process items that have already been created
                        if itemname in self.Items:

                            item = self.Items[itemname]

                            if command == 'IMAGE':

                                if item.ItemType == IT_IMAGE:

                                    # Assign a new filename for an image item
                                    if sh.cell(rowx=Row,colx=Col).ctype==xlrd.XL_CELL_TEXT:

                                        filename = os.path.join(self.ImageDir, sh.cell(rowx=Row,colx=Col).value)

                                        assert os.path.isfile(filename), \
                                            "%s at row %d could not find file %s" % (command, Row+1, filename)

                                        self.Items[itemname].AddImage( time, filename )

                            elif command == 'MASK':

                                if self.Items[item].ItemType == IT_MASK:
                                    # Assign a new filename for an image item
                                    if sh.cell(rowx=Row,colx=Col).ctype==xlrd.XL_CELL_TEXT:

                                        filename = os.path.join(self.ImageDir, sh.cell(rowx=Row,colx=3).value)

                                        assert os.path.isfile(filename), \
                                            "%s at row %d could not find file %s" % (command, Row+1, filename)

                                        self.Items[itemname].AddMask( time, filename )

                            elif command == 'TEXT':

                                if item.ItemType == IT_TEXT:

                                    # Assign a new title for a text item
                                    if sh.cell(rowx=Row,colx=Col).ctype==xlrd.XL_CELL_TEXT:

                                        text = sh.cell(rowx=Row,colx=Col).value
                                        self.Items[itemname].AddText( time, text )

                            elif command in ['XPOS', 'YPOS', 'XPOLE', 'YPOLE', 'XSCALE', 'YSCALE', 'ROTATION', 
                                             'TIMEOFFSET', 'TEXTSIZE', 'OPACITY']:

                                # Set a new float property
                                if sh.cell(rowx=Row,colx=Col).ctype==xlrd.XL_CELL_NUMBER:
                                    xpos = sh.cell(rowx=Row,colx=Col).value
                                    self.Items[itemname].Properties[command].Append(time, xpos)

                            elif command in ['XMOVE', 'YMOVE', 'SXMOVE', 'SYMOVE', 'RMOVE']:
                                # Determine type of movement
                                if sh.cell(rowx=Row,colx=Col).ctype==xlrd.XL_CELL_TEXT:
                                    move = sh.cell(rowx=Row,colx=Col).value.strip().upper()
                                    if move in CheckMove:
                                        self.Items[itemname].Properties[command].Append(time, CheckMove[move])
                                    else:
                                        print("Did not recognize type of movement on row %d." % (Row+1))


                            elif command in ['TEXTCOLOR', 'FONT']:
                                if sh.cell(rowx=Row,colx=Col).ctype==xlrd.XL_CELL_TEXT:
                                    textcolor = sh.cell(rowx=Row,colx=Col).value.strip()
                                    self.Items[itemname].Properties[command].Append(time, textcolor)


                            else:
                                print('Command: ', command)
                                print('Column: ', Col+1)
                                print("Command %s not recognized on col %d." % (command, Col+1))

    def StandardChecks(self):

        print(' - checking script %s which has %d items' % (self.SheetName, len(self.Items) ))

        # Do some standard checks after parsing
        OK = True
        self.TimeOffsetUsed=False
        
        for i in self.Items.values():
            
            i.StandardChecks()

            if i.TimeOffsetUsed:
                self.TimeOffsetUsed=True

            if (i.ItemType == IT_IMAGE):

                if len(i.Properties['IMAGE'].Sequence)==0:

                    print('ERROR: %s has NO images' % i.ItemName)
                    OK=False

                else:

                    for time, filename in i.Properties['IMAGE'].Sequence:

                        if not os.path.isfile(filename):
                            print('Image not found: %s at tim %.3f' % (filename, time))
                            OK = False

            if (i.ItemType == IT_MASK):

                if len(i.Properties['MASK'].Sequence)==0:

                    print('ERROR: %s has NO mask' % i.ItemName)
                    OK=False

                else:

                    for time, filename in i.Properties['MASK'].Sequence:

                        if not os.path.isfile(filename):
                            print('Mask not found: %s at tim %.3f' % (filename, time))
                            OK = False

            if (i.ItemType == IT_TEXT):

                if len(i.Properties['TEXT'].Sequence)==0:
                    print('ERROR: %s has NO lines of text' % i.ItemName)
                    OK=False

        return OK


    def Deploy(self, MaxTime):

        logging.debug('')
        logging.debug('* DEPLOYING SCRIPT %s' % self.SheetName)

        for item in self.Items.values():
            item.Deploy(MaxTime)
            
        if not self.Zbuffer:
            # The Zbuffer has no items because the user did not specify 
            # any BRINGTOFRONT or SENDTOBACK commands
            
            # Get the name of a random item
            itemname = list(self.Items.keys())[0]
            
            self.Zbuffer.append( ( 0, itemname, FRONT) )

        self.Zbuffer.sort()

        time, item, direction = self.Zbuffer[-1]
        self.Zbuffer.append( (MaxTime, item, direction) )
        self.Zbuffer.sort()

        # Determine the order of the items at time = 0
        self.ZbufferIndex = 0

        # list() means we create a copy
        self.Zorder = list(self.Items.keys())


    def GetPicture(self, Time, Frame):

        # If exactly the same image was calculated before,
        # use that image
        #if Frame != self.PictureFrame and not self.TimeOffsetUsed:
        if True:

            logging.debug('')
            logging.debug('* SCRIPT %s IS GENERATING FRAME %.5d at time %.2f' % (self.SheetName, Frame, Time ))            

            # Start with a transparent image
            if (not self.IsCanvas) or self.FirstImage:
                self.Picture = Image.new("RGBA", (self.Width, self.Height), (255,0,0,0) )

            self.FirstImage=False

            # Determine the Z-order at the desired time
            while True:

                time, item, direction = self.Zbuffer[self.ZbufferIndex]

                if item not in self.Zorder:
                    print('Z-order failure: item %s not in script %s' % (item, self.SheetName) )

                self.Zorder.remove(item)

                if direction == FRONT:
                    self.Zorder.append(item)
                else:
                    self.Zorder.insert(0, item)

                if (self.Zbuffer[self.ZbufferIndex+1][0])>Time:
                    break
                else:
                    self.ZbufferIndex+=1

                
            ItemPicture = Image.new("RGBA", (self.Width, self.Height), (255,0,0,0) )

            # Draw each item
            for itemname in self.Zorder:

                Item = self.Items[itemname]

                opacity = Item.Properties['OPACITY' ].Value(Time, LINEAR)
                move    = Item.Properties['XMOVE'   ].Value(Time)
                xpos    = Item.Properties['XPOS'    ].Value(Time, move)
                move    = Item.Properties['YMOVE'   ].Value(Time)
                ypos    = Item.Properties['YPOS'    ].Value(Time, move)
                move    = Item.Properties['SXMOVE'  ].Value(Time)
                sx      = Item.Properties['XSCALE'  ].Value(Time, move)
                move    = Item.Properties['SYMOVE'  ].Value(Time)
                sy      = Item.Properties['YSCALE'  ].Value(Time, move)
                move    = Item.Properties['RMOVE'   ].Value(Time)
                rot     = Item.Properties['ROTATION'].Value(Time, move)

                try:
                    logging.debug('  - Item %s:%s xpos= %.2f ypos= %.2f xscale= %.3f yscale= %.3f rot= %.3f opacity= %.3f' % (self.SheetName, itemname, xpos, ypos, sx, sy, rot, opacity))
                except:
                    print('opacity', opacity)
                    print('xpos', xpos)
                    print('ypos', ypos)
                    print('sx', sx)
                    print('sy', sy)
                    print('rot', rot)

                if opacity>0:

                    if Item.ItemType == IT_ASSY:

                        script = Item.Properties['SCRIPT'].Value(Time)
                        logging.debug('  - Assembly %s:%s requests an image from script %s' % (self.SheetName, itemname, script))

                        if script in self.ScriptList:
                            dt=Item.Properties['TIMEOFFSET'].Value(Time, LINEAR)
                            ItemPicture = self.ScriptList[script].GetPicture(Time-dt, Frame)
                        else:
                            logging.debug('  Script %s not in scriptlist!!:'% (script))
                            ItemPicture = Image.new("RGBA", (self.Width, self.Height), (255,0,0,0) )
                            
                        logging.debug('  Assembly %s continues:'% (self.SheetName))

                    if Item.ItemType == IT_CANVAS:

                        script = Item.Properties['SCRIPT'].Value(Time)
                        logging.debug('  - Canvas %s:%s requests an image from script %s' % (self.SheetName, itemname, script))

                        if script in self.ScriptList:
                            dt=Item.Properties['TIMEOFFSET'].Value(Time, LINEAR)
                            ItemPicture = self.ScriptList[script].GetPicture(Time-dt, Frame)
                        else:
                            ItemPicture = Image.new("RGBA", (self.Width, self.Height), (255,0,0,0) )

                    elif Item.ItemType == IT_IMAGE:

                        image = Item.Properties['IMAGE'].Value(Time)

                        if Item.PrevImageName != image:
                            Item.LoadedImage = Image.open(image).convert("RGBA")
                            Item.PrevImageName = image

                        ItemPicture = Item.LoadedImage


                    elif Item.ItemType == IT_MASK:

                        image = Item.Properties['MASK'].Value(Time)
                        logging.debug('Line 585 mask is %s' % image)

                        if Item.PrevImageName != image:
                            Item.LoadedImage = Image.open(image).convert("RGBA")
                            Item.PrevImageName = image

                        ItemPicture = Item.LoadedImage

                    elif Item.ItemType == IT_TEXT:

                        ItemPicture = Image.new("RGBA", (self.Width, self.Height), (255,0,0,0) )

                        text      = Item.Properties['TEXT'     ].Value(Time)
                        textsize  = int(Item.Properties['TEXTSIZE' ].Value(Time, LINEAR))
                        textcolor = Item.Properties['TEXTCOLOR'].Value(Time)
                        fontname  = Item.Properties['FONT'     ].Value(Time)

                        Directories = [ 'C:\\WINDOWS\\Fonts\\' ]
                        Fonts = [fontname, 'calibri.ttf', 'YanoneKaffeesatz-Regular.ttf', 'ARIALN.TTF', 'verdana.ttf', 'YanoneKaffeesatz-Light.ttf']

                        Face = ImageFont.truetype(SelectFont(Directories, Fonts), textsize)
                        Draw = ImageDraw.Draw(ItemPicture)
                        Draw.text( (0,0), text, fill=textcolor, font=Face)

                    # Retrieve the general properties
                    move  = Item.Properties['XMOVE'   ].Value(Time)
                    xpos  = Item.Properties['XPOS'    ].Value(Time, move)

                    move  = Item.Properties['YMOVE'   ].Value(Time)
                    ypos  = Item.Properties['YPOS'    ].Value(Time, move)

                    move  = Item.Properties['SXMOVE'  ].Value(Time)
                    sx    = Item.Properties['XSCALE'  ].Value(Time, move)
                    xpole = Item.Properties['XPOLE'   ].Value(Time, move)

                    move  = Item.Properties['SYMOVE'  ].Value(Time)
                    sy    = Item.Properties['YSCALE'  ].Value(Time, move)
                    ypole = Item.Properties['YPOLE'   ].Value(Time, move)

                    move  = Item.Properties['RMOVE'   ].Value(Time)
                    rot   = Item.Properties['ROTATION'].Value(Time, move)

                    fi    = math.pi/180*rot
                    sinfi = math.sin(fi)
                    cosfi = math.cos(fi)

                    w,h = ItemPicture.size

                    # Resize and rotate the ItemPicture
                    try:
                        ItemPicture=ItemPicture.resize( (int(sx*w+0.5), int(sy*h+0.5) ), Image.ANTIALIAS)
                        ItemPicture=ItemPicture.rotate(rot, expand=1)
                    except:
                        print('ERROR Script 663: Item %s:%s sx= %.2f sy= %.2f' % (self.SheetName, itemname, sx, sy))
                        break

                    wr,hr = ItemPicture.size

                    xt = xpos + xpole - ypole*sy*sinfi - xpole*sx*cosfi +0.5*w*sx*cosfi +0.5*h*sy*sinfi -0.5*wr
                    yt = ypos + ypole - ypole*sy*cosfi + xpole*sx*sinfi -0.5*w*sx*sinfi +0.5*h*sy*cosfi -0.5*hr

                    Mask = ItemPicture.convert("RGBA")
                    Mask = Image.blend(Image.new(ItemPicture.mode, ItemPicture.size, 0), ItemPicture, opacity)

                    if Item.ItemType != IT_MASK:

                        # Item is picture, assembly or canvas
                        self.Picture.paste( ItemPicture, (int(xt),int(yt)), Mask )

                    else:

                        # Item is mask
                        logging.debug('  - Applying mask for %s' % itemname)

                        # Start with a clean image with transparent background
                        CleanImage = Image.new("RGBA", (self.Width, self.Height), (0,0,0,0) )

                        # Use the mask rotated and translated
                        Mask = Image.new("L", (self.Width, self.Height), 0 )
                        Mask.paste( ItemPicture, (int(xt),int(yt)))

                        # Copy the image as-is with rotation and translation set to zero
                        CleanImage.paste( self.Picture, (0,0), Mask )

                        self.Picture = CleanImage.copy()

            self.PictureFrame = Frame

        return self.Picture.copy()
